#! /usr/bin/env python

import xmlrpclib
from peerDescriptor import PeerDescriptor
from persistentDict import PersistentDict
from filesMetadata import serializeFileInfo
from utils import lock
import re
import logs
import random
import socket
from threading import Thread
from common import day_difference, calcResiliencyLevel
import time
import os

class PingAgent(Thread):
    def __init__(self, dhtAddr, host, ttlMap, rttMap, peerId):
        Thread.__init__(self)
        self.dhtAddr = dhtAddr      
        self.host = host
        self.ttlMap = ttlMap
        self.rttMap = rttMap
        self.peerId = peerId

    def markAvailability(self, val):
        pass
        #proxy = xmlrpclib.ServerProxy(self.dhtAddr, allow_none=True)
        #proxy.markAvailability(str(self.peerId), val)

    def run(self):
        out = os.popen('ping -c 1 ' + str(self.host))
        output = out.read()
        rttS = re.search('time=([0-9]*)', output)
        ttlS = re.search('ttl=([0-9]*)', output)
        if rttS:
            rtt = rttS.group(1)
            self.rttMap.set([self.peerId], rtt)
        if ttlS:
            ttl = int(ttlS.group(1))
            if ttl <= 64:
                dist = 64 - ttl
            elif ttl <= 128:
                dist = 128 - ttl
            else:
                dist = 256 - ttl
            self.ttlMap.set([self.peerId], dist)
        if rttS != None and ttlS != None:
            self.markAvailability(1.0)
        else:
            self.markAvailability(0.0)

class PeersMonitor():
    def __init__(self, serverHttpAddr, myId, config, myPortNum, metadataPath):
        self.myId = myId
        self.config = config
        self.serverHttpAddr = serverHttpAddr
        self.ttlMap = PersistentDict(metadataPath + "/ttlMap-" + str(myPortNum))
        self.rttMap = PersistentDict(metadataPath+ "/rttMap-" + str(myPortNum))
        self.performanceInfoMap = PersistentDict(metadataPath + "/performanceInfoMap-" + str(myPortNum))

    def sync(self):
        self.ttlMap.sync()
        self.rttMap.sync()

    def getTtl(self, peerId):
        try:
            res = self.ttlMap.get([peerId])
        except KeyError:
            res = 0
        return res

    def getRtt(self, peerId):
        try:
            res = self.rttMap.get([peerId])
        except KeyError:
            res = None
        return res

    def getPerformanceInfo(self, peerId):
        assert self.performanceInfoMap.has_key([peerId])
        res = PerformanceInfo(peerId)
        perfInfoDict = self.performanceInfoMap.get([peerId])
        res.totalIncommingSize = int(perfInfoDict['totalIncommingSize'])
        res.totalIncommingBwBound = int(perfInfoDict['totalIncommingBwBound'])
        res.availability = perfInfoDict['availability']

        assert(res.totalIncommingSize != None)
        assert(res.totalIncommingBwBound != None)

        return res

    def setPerformanceInfo(self, perfInfo, ver):
        assert(perfInfo.totalIncommingSize != None)
        assert(perfInfo.totalIncommingBwBound != None)

        if not self.performanceInfoMap.has_key([perfInfo.peerId]):
            peerInfoMap = {'peerId': perfInfo.peerId, 'version': -2}
        else:
            peerInfoMap = self.performanceInfoMap.get([perfInfo.peerId])

        if ver > peerInfoMap['version']:
            if perfInfo.totalIncommingSize != None:
                peerInfoMap['totalIncommingSize'] = str(perfInfo.totalIncommingSize)
            if perfInfo.totalIncommingBwBound != None:
                peerInfoMap['totalIncommingBwBound'] = str(perfInfo.totalIncommingBwBound)
            if perfInfo.availability != None:
                peerInfoMap['availability'] = perfInfo.availability
            self.performanceInfoMap.set([perfInfo.peerId], peerInfoMap)
        self.performanceInfoMap.sync()

    def printReport(self):
        print(str(self.myId) + ":")
        allKeys = self.performanceInfoMap.keys([])
        #for key in self.rttMap.keys():
        #    if key not in allKeys:
        #        allKeys.append(key)
        #for key in self.ttlMap.keys():
        #    if key not in allKeys:
        #        allKeys.append(key)
        for key in allKeys:
            print("    " + str(key) + " = (rtt = " + str(self.getRtt(key)) + ", ttl = " + str(self.getTtl(key)) + ", perfInfo = " + str(self.performanceInfoMap.get([key])) + ")")

    def run(self):
        proxy = xmlrpclib.ServerProxy(self.serverHttpAddr, allow_none=True)
        while True:
            try:
                allPeerIds = proxy.getAllPeerIds()
                di = proxy.getDict()
                current = time.time()
                for peer in allPeerIds:
                    obj, ver = di[peer]
                    if not self.performanceInfoMap.has_key(peer):
                        peerInfoMap = {'peerId': peer, 'version': -2}
                    else:
                        peerInfoMap = self.performanceInfoMap.get([peer])

                    perMapVer = peerInfoMap['version']
                    if (obj["ipAddress"] != None):
                        pa = PingAgent(self.serverHttpAddr, obj["ipAddress"], self.ttlMap, self.rttMap, peer)
                        pa.start()

                    if ver > perMapVer:
                        if (obj["availabilityTimestamp"] == None) or (obj["availabilityMeasurementDuration"] == None) or (obj["availability"] == None):
                            peerInfoMap['availability'] = None
                        else:
                            unavDur = max(0.0, current -  obj["availabilityTimestamp"])
                            avDur = obj["availabilityMeasurementDuration"]
                            av = float(obj["availability"])
                            peerInfoMap['availability'] = av * (1 - unavDur / (unavDur + avDur))
                        totalIncommingSize = obj['totalIncommingSize']
                        if totalIncommingSize != None:
                            totalIncommingSize = int(totalIncommingSize)
                        peerInfoMap['totalIncommingSize'] = str(totalIncommingSize)
                        totalIncommingBwBound = obj['totalIncommingBwBound']
                        if totalIncommingBwBound != None:
                            totalIncommingBwBound = int(totalIncommingBwBound)
                        peerInfoMap['totalIncommingBwBound'] = str(totalIncommingBwBound)
                        peerInfoMap['version'] = ver
                        self.performanceInfoMap.set([peer], peerInfoMap)
                self.performanceInfoMap.sync()
                #self.printReport()
            except Exception as e:
                logs.trace(e)
                pass # ignorring exceptions
            self.sync()
            time.sleep(self.config.peersManagerIntervalSec)


class NotEnoughInformation(Exception):
    pass

class RemoteLocation:
    def __init__(self, serverHttpAddrArg, peerIdArg, avArg, myAvArg, totalIncommingSize, totalIncommingBwBound, desiredBackWindow, ttl = None):
        self.serverHttpAddr = serverHttpAddrArg
        self.peerId = peerIdArg
        self.myAvailability = float(myAvArg)
        if ttl == None:
            self.ttlDist = None
        else:
            self.ttlDist = int(ttl)
        self.availability = float(avArg)
        self.totalIncommingSize = int(totalIncommingSize) # bytes
        self.totalIncommingBwBound = int(totalIncommingBwBound)
        self.desiredBackWindow = int(desiredBackWindow) # seconds

    def expectedDataTransferDuration(self):
        if (self.myAvailability < 0.01) or (self.availability < 0.01):
            return 60 * 60 * 24 * 365 # very large number
        av = self.availability;
        time = self.totalIncommingSize / self.totalIncommingBwBound
        return av * time

    def perfScore(self):
        return min(self.desiredBackWindow - self.expectedDataTransferDuration(), 0)

    def __str__(self):
        # + ", ttl = " + str(self.ttlDist) + ", av = " + str(self.availability) + "," +
        return "Peer: " + str(self.peerId) + \
               " incSize = " + str(self.totalIncommingSize) + ", bwBound = " + str(self.totalIncommingBwBound)

class ReplicaPlacement:
    def __init__(self, ownerId, fileName, config):
        self.remoteLocations = []
        self.ownerId = ownerId
        self.resiliencyLevel = calcResiliencyLevel(filename = fileName)
        self.config = config
        self.lowTTLmax = 1
        self.highTTLmin = 5
        self.highTTLmax = 15

    def addPeer(self, remoteLocation):
        for location in self.remoteLocations:
            if location.peerId == remoteLocation.peerId:
                return
        self.remoteLocations.append(remoteLocation)

    def removePeer(self, remoteLocation):
        for location in self.remoteLocations:
            if str(location.peerId) == str(remoteLocation.peerId):
                self.remoteLocations.remove(location)

    def getTTLScore(self):
        ttls = []
        for loc in self.remoteLocations:
            ttls.append(loc.ttlDist)
        ttls.sort()
        if len(ttls) == 0:
            return 0
        high = ttls.pop()
        error = 0
        if high < self.highTTLmin:
            error += (self.highTTLmin - high)
        if high > self.highTTLmax:
            error += (high - self.highTTLmax)
        for low in ttls:
            if low > self.lowTTLmax:
                error += (low - self.lowTTLmax)
        return -1 * error
        
    def getResiliencyScore(self):
        error = abs(self.resiliencyLevel - len(self.remoteLocations))
        return -100000 * error

    def getPerfScore(self):
        score = 0.
        for loc in self.remoteLocations:
            score += loc.perfScore()
        return int (score / self.config.accuracyOfPerformanceScore) # scores have some accuracy - they are discretitized

    def score(self):
        return self.getTTLScore() + self.getPerfScore() + self.getResiliencyScore()


# predefined replicas that constitute scores

class PerformanceInfo:
    def __init__(self, peerId):
        self.peerId = str(peerId)
        self.totalIncommingSize = 0
        self.totalIncommingBwBound = 0
        self.availability = 0.0

    def __repr__(self):
        return "<IncSize=" + str(self.totalIncommingSize) + ", BwBound=" + str(self.totalIncommingBwBound) + ", av=" + str(self.availability) + ">"

def perfInfoFromDict(di):
    peerId = di['peerId']
    perfInfo = PerformanceInfo(peerId)
    perfInfo.availability = float(di['availability'])
    perfInfo.totalIncommingSize = int(di['totalIncommingSize'])
    perfInfo.totalIncommingBwBound = int(di['totalIncommingBwBound'])
    return perfInfo

class FilesLocationsManager:
    def __init__(self, serverHttpAddrArg, metadataPath, availabilityManager, filesOperations, myAddr, myPortNum, myId, config, perfParams):
        self.serverHttpAddr = serverHttpAddrArg
        self.config = config
        self.peersMonitor = PeersMonitor(self.serverHttpAddr, myId, config, myPortNum, metadataPath)
        self.filesOperations = filesOperations
        self.myPeerId = self.filesOperations.myPeerId
        self.lockfile = lock(os.path.join(self.filesOperations.basicPath, "filesLocationsManagerLock"))
        self.scores = {} # this is the map that for each fileId holds a pair (score and size)
        self.availabilityManager = None
        self.myAddr = myAddr
        self.commiting = False
        self.myPortNum = myPortNum
        self.perfParams = perfParams
        self.perfInfo = PerformanceInfo(self.myPeerId)
        self.availabilityManager = availabilityManager

    def setTotalIncommingBwBound(self, val):
        self.perfParams.totalIncommingBwBound = val

    def preparePerfromanceInfo(self):
        myProxy = xmlrpclib.ServerProxy("http://"+str(self.myAddr) + ":" + str(self.myPortNum), allow_none = True)
        agreementsSize = int(myProxy.getAgreementsSize())
        self.perfInfo.totalIncommingSize = str(agreementsSize)
        self.perfInfo.totalIncommingBwBound = str(self.perfParams.totalIncommingBwBound)
        self.perfInfo.availability = self.availabilityManager.getAvailability()
        
    def updatePerformanceInfo(self, perfInfoDict, ver):
        perfInfo = perfInfoFromDict(perfInfoDict)
        self.peersMonitor.setPerformanceInfo(perfInfo, ver)

    def periodicMeasurement(self):
        self.peersMonitor.run()

    def tryCommitContracts(self):
        self.commiting = True
        start_time = time.time()
        try:
            myProxy = xmlrpclib.ServerProxy("http://"+str(self.myAddr) + ":" + str(self.myPortNum), allow_none = True)
            allFilesInfo = self.filesOperations.filesCatalog.getAllFileInfo()

            toSentReplPerOwner = {}
            for fInfo in allFilesInfo:
                while len(fInfo.committedLocation) < len(fInfo.replicatorsIds):
                    fInfo.committedLocation.append((None, 0, True))
                while len(fInfo.oldLocations) < len(fInfo.replicatorsIds):
                    fInfo.oldLocations.append(None)
                for i in range(0, len(fInfo.committedLocation)):
                    loc, t, sent = fInfo.committedLocation[i]
                    oldLoc = loc
                    if day_difference(t, time.time()) > 0:
                        sent = False
                        try:
                            loc = fInfo.replicatorsIds[i]
                        except IndexError:
                            continue
                        t = time.time()
                        fInfo.committedLocation[i] = (loc, t, sent)
                        logs.debug("TMAN_SERVER: Commited file: " + fInfo.lastLocation + " in loaction: " + str(loc))
                        if (oldLoc != fInfo.oldLocations[i]) and (oldLoc != None):
                            logs.debug("TMAN_SERVER: Removing file: " + fInfo.lastLocation + " from location which did not ack : " + str(oldLoc))
                            myProxy.addToRemove(oldLoc, "unknown", fInfo.lastLocation)
                    if sent == False:
                        try:
                            toSentReplPerOwner[loc].append((fInfo, i))
                        except KeyError:
                            toSentReplPerOwner[loc] = [(fInfo, i)]
                self.filesOperations.filesCatalog.put(fInfo)
            #logs.debug("TMAN_SERVER: PERFORMANCE: First phase of commit duration: " + str(time.time() - start_time))
            toSentKeys = toSentReplPerOwner.keys()
            random.shuffle(toSentKeys)
            for loc in toSentKeys:
                try:
                    inter_time_start = time.time()
                    myProxy.removeFromToRemove(
                        loc,
                        [fInfo.lastLocation for (fInfo, i) in toSentReplPerOwner[loc]]
                    )
                   # logs.debug("TMAN_SERVER: PERFORMANCE: Commit inter time duration 1: " + str(time.time() - inter_time_start))
                    proposedPeer = PeerDescriptor(self.serverHttpAddr, loc)
                    proposedPeerNegoProxy = xmlrpclib.ServerProxy(
                                               "http://"+str(proposedPeer.ipAddress) + ":" + str(proposedPeer.portNumber),
                                                allow_none = True
                                            )
                    proposedPeerNegoProxy.replicateFilesFrom(
                        self.myPeerId,
                        [fInfo for (fInfo, i) in toSentReplPerOwner[loc]]
                    )
                    #logs.debug("TMAN_SERVER: PERFORMANCE: Commit inter time duration 2: " + str(time.time() - inter_time_start))
                    for (fInfo, i) in toSentReplPerOwner[loc]:
                        fInfo.committedLocation[i] = (loc, t, True)
                        logs.debug("TMAN_SERVER: File: " + fInfo.lastLocation + " ordered to be replicated at: " + str(loc))
                        self.filesOperations.filesCatalog.put(fInfo)
                    #logs.debug("TMAN_SERVER: PERFORMANCE: Commit inter time duration total: " + str(time.time() - inter_time_start))
                except socket.error as e:
                    logs.commonErr(e)
                except Exception as e:
                    logs.trace(e)
            #logs.debug("TMAN_SERVER: PERFORMANCE: Second phase of commit duration: " + str(time.time() - start_time))
            self.filesOperations.filesCatalog.sync()
        except Exception as e:
            logs.trace(e)
        self.commiting = False

    def updateScores(self):
        allFilesInfo = self.filesOperations.filesCatalog.getAllFileInfo()
        for i in allFilesInfo:
            self.updateScore(i)

    def updateScore(self, fileInfo):
        try:
            self.scores[fileInfo.fileId] = (self.scoreLocations(fileInfo.replicatorsIds, self.myPeerId, fileInfo.lastLocation), fileInfo.size)
        except NotEnoughInformation:
            pass

    def getAvailability(self):
        return self.perfInfo.availability

    def getTtls(self):
        result = {}
        result['far'] = {}
        result['close'] = {}
        try:
            allFilesInfo = self.filesOperations.filesCatalog.getAllFileInfo()
            for fInfo in allFilesInfo:
                vals = []
                for rep in fInfo.replicatorsIds:
                    vals.append(self.peersMonitor.getTtl(rep))
                vals.sort()
                for i in range(0, len(vals) - 1):
                    v = vals[i]
                    try:
                        result['close'][str(v)] += 1
                    except KeyError:
                        result['close'][str(v)] = 1
                if len(vals) > 0:
                    v = vals[-1]
                    try:
                        result['far'][str(v)] += 1
                    except KeyError:
                        result['far'][str(v)] = 1
        except Exception as e:
            logs.trace(e)
        logs.debug("TMAN_SERVER: TTLs returned: " + str(result))
        return result       

    def proposedReplication(self, peerId, peerUsername, offeredSize):
        try:
            logs.debug("TMAN_SERVER: Got proposal from peer " + str(peerId) + " (" + peerUsername + ")")
            myProxy = xmlrpclib.ServerProxy("http://"+str(self.myAddr) + ":" + str(self.myPortNum), allow_none = True)
            if self.filesOperations.filesCatalog == None:
                raise Exception("No files cataloge. Try later")
            scores = self.scores.items()
            allFilesInfo = self.filesOperations.filesCatalog.getStorageMap()
            acceptedNoRejectedSize = 0
            acceptedNoRejectedFiles = []
            rejectedSizes = {} # map from peerId to the rejected size
            rejectedFiles = {} # map from peerId to the list of rejected files
            rejectedImpr = {} # map from peerId to the value of improvement
            totalIncommingSizes = {} # map from peerId to the value of improvement

            proposedPeer = PeerDescriptor(self.serverHttpAddr, peerId)
            proposedPeerNegoProxy = xmlrpclib.ServerProxy("http://"+str(proposedPeer.ipAddress) + ":" + str(proposedPeer.portNumber), allow_none = True)
            peerIncommingSize = self.peersMonitor.getPerformanceInfo(peerId).totalIncommingSize

            for fileId, (s, size) in sorted(scores, key=lambda x: x[1]):
                fileInfo = allFilesInfo[fileId]
                curReplicators = fileInfo.replicatorsIds
                if peerId in curReplicators:
                    logs.debug("TMAN_SERVER: File " + fileInfo.lastLocation + " is already in replicators")
                    continue
                if calcResiliencyLevel(filename = fileInfo.lastLocation) > len(curReplicators):
                    if acceptedNoRejectedSize + int(fileInfo.size) <= offeredSize:
                        acceptedNoRejectedSize += int(fileInfo.size)
                        acceptedNoRejectedFiles.append(fileId)
                if acceptedNoRejectedSize == 0: # only if all files have sufficient resiliency level we try to change replicators
                    try:
                        scoresImpr = self.scoreEachLocationExchange(
                                          locations = curReplicators,
                                          newLoc = peerId,
                                          newLocIncommingSize = peerIncommingSize,
                                          fileName = fileInfo.lastLocation,
                                          alreadyRejected = rejectedSizes,
                                          totalIncommingSizes = totalIncommingSizes)
                    except NotEnoughInformation:
                        logs.debug("TMAN_SERVER: Not enough information about peer")
                        continue
                    for newPeerId, scoreGain in scoresImpr.items():
                        try:
                            alreadyAcc = rejectedSizes[newPeerId]
                        except KeyError:
                            alreadyAcc = 0
                        if alreadyAcc + int(fileInfo.size) > offeredSize:
                            logs.debug("TMAN_SERVER: New peer has not enough size alreadyAcc: " + str(alreadyAcc) + \
                                       ", file size: " + str(fileInfo.size) + ", offeredSize: " + str(offeredSize))
                        if (scoreGain < self.config.scoreGainThreshold):
                            logs.debug("TMAN_SERVER: Peer " + str(newPeerId) + " changing below the threshold: scoreGain: " + \
                                       str(scoreGain) + ", threshold: " + str(self.config.scoreGainThreshold)) 
                        if (scoreGain > self.config.scoreGainThreshold) and (alreadyAcc + int(fileInfo.size) <= offeredSize):
                            try:
                                rejectedImpr[newPeerId] = max(rejectedImpr[newPeerId], scoreGain)
                            except KeyError:
                                rejectedImpr[newPeerId] = scoreGain
                            try:
                                rejectedSizes[newPeerId] += int(fileInfo.size)
                            except KeyError:
                                rejectedSizes[newPeerId] = int(fileInfo.size)
                            try:
                                rejectedFiles[newPeerId].append(fileId)
                            except KeyError:
                                rejectedFiles[newPeerId] = [fileId]

            if acceptedNoRejectedSize > 0:
                accFiles = []
                for fileId in acceptedNoRejectedFiles:
                    fileInfo = allFilesInfo[fileId]
                    accFiles.append((fileInfo.lastLocation, fileInfo.size))
                    self.filesOperations.changeLocationInFilesCatalog(fInfo = fileInfo, location = peerId, toRemove = None)
                logs.info("TMAN_SERVER: Peer accepted replica invitation from " + str(peerId) + " as we have too little replicas")
                proposedPeerNegoProxy.acceptStorageContract(self.myPeerId, str(acceptedNoRejectedSize), accFiles)

                return (True, False) # accepted invitation and perf info has not changed
            else:
                perfInfoChanged = False
                for toRemove, impr in sorted(rejectedImpr.items(), key=lambda x: -1 * x[1]):
                    logs.debug("TMAN_SERVER: Trying to change new peer: " + str(peerId) +  " for peer: " + str(toRemove))
                    try:
                        if self.peersMonitor.getPerformanceInfo(toRemove).availability < \
                           self.config.availabilityThresholdForAsynchronousRevokingContract:

                            logs.info("TMAN_SERVER: Peer availability is too low. Therefore we reject the contract even when not available.")
                            result = proposedPeerNegoProxy.acceptStorageContractIf(
                                         self.myPeerId,
                                         str(rejectedSizes[toRemove]),
                                         str(peerIncommingSize),
                                         [(allFilesInfo[fileId].lastLocation, allFilesInfo[fileId].size) for fileId in rejectedFiles[toRemove]]
                                     )
                            if result == 0:
                                perfInfoChanged = True
                                logs.debug("TMAN_SERVER: Performance info of proposing peer " + str(peerId) + " has changed - resigned from proporsal")
                                break

                            for fileId in rejectedFiles[toRemove]:
                                fileInfo = allFilesInfo[fileId]
                                self.filesOperations.changeLocationInFilesCatalog(fileInfo, peerId, toRemove)

                            logs.info("TMAN_SERVER: Exchanged peer " + str(toRemove) + " for peer " + str(peerId))
                            logs.debug("TMAN_SERVER: Size of exchanged files " + str(rejectedSizes[toRemove]))
                            logs.debug("TMAN_SERVER: List of exchanged files " + str(rejectedFiles[toRemove]))

                            return (True, False)

                        else:

                            remotePeer = PeerDescriptor(self.serverHttpAddr, toRemove)
                            negoProxy = xmlrpclib.ServerProxy("http://"+str(remotePeer.ipAddress) + ":" + str(remotePeer.portNumber), allow_none = True)
                            result = negoProxy.rejectStorageContractIf(
                                         self.myPeerId,
                                         str(rejectedSizes[toRemove]),
                                         [(allFilesInfo[f].lastLocation, str(allFilesInfo[f].size)) for f in rejectedFiles[toRemove]],
                                         str(totalIncommingSizes[toRemove])
                                     )
                            if result == 0:
                                perfInfoChanged = True
                                logs.debug("TMAN_SERVER: Performance info of peer to remove " + str(toRemove) + " has changed - resigned from proporsal")
                                continue
                            result = proposedPeerNegoProxy.acceptStorageContractIf(
                                         self.myPeerId,
                                         str(rejectedSizes[toRemove]),
                                         str(peerIncommingSize),
                                         [(allFilesInfo[fileId].lastLocation, allFilesInfo[fileId].size) for fileId in rejectedFiles[toRemove]]
                                     )
                            if result == 0:
                                negoProxy.acceptStorageContract(
                                    self.myPeerId,
                                    str(rejectedSizes[toRemove]),
                                    [(allFilesInfo[fileId].lastLocation, allFilesInfo[fileId].size) for fileId in rejectedFiles[toRemove]]
                                )
                                perfInfoChanged = True
                                logs.debug("TMAN_SERVER: Performance info of proposing peer " + str(peerId) + " has changed - resigned from proporsal")
                                continue

                            for fileId in rejectedFiles[toRemove]:
                                fileInfo = allFilesInfo[fileId]
                                self.filesOperations.changeLocationInFilesCatalog(fileInfo, peerId, toRemove)

                            logs.info("TMAN_SERVER: Exchanged peer " + str(toRemove) +  " (" + remotePeer.username + ") for peer " + \
                                      str(peerId) + " (" + str(peerUsername) + ")")
                            logs.debug("TMAN_SERVER: Size of exchanged files " + str(rejectedSizes[toRemove]))
                            logs.debug("TMAN_SERVER: List of exchanged files " + str(rejectedFiles[toRemove])) 
                            return (True, False)
                    except socket.error as e:
                        logs.commonErr(e)
                        logs.info("TMAN_SERVER: Could not exchange contract because peer " + str(peerId) + " is unavailable, error: " + str(e))
                    except Exception as e:
                        logs.trace(e)

                logs.info("TMAN_SERVER: Rejected proposal of peer " + str(peerId))
                return (False, perfInfoChanged)
        except Exception as e:
            logs.trace(e)
            raise e

    def ackDownload(self, filename, peerId, peerUsername):
        myProxy = xmlrpclib.ServerProxy("http://"+str(self.myAddr) + ":" + str(self.myPortNum), allow_none = True)
        fInfo = self.filesOperations.filesCatalog.find(filename)
        try:
            ind = -1
            for l in fInfo.committedLocation:
                if l[0] == peerId:
                    ind = fInfo.committedLocation.index(l)
            if ind == -1:
                logs.debug("TMAN_SERVER: Peer " + peerId + " has been previously removed from commitedPeers for file: " + str(filename))
                if peerId not in fInfo.oldLocations:
                    myProxy.addToRemove(peerId, peerUsername, filename)
                return
            oldLoc = fInfo.oldLocations[ind]
            if (oldLoc != peerId) and (oldLoc != None):
                myProxy.addToRemove(oldLoc, peerUsername, filename)
                fInfo.oldLocations[ind] = peerId
              
            self.filesOperations.filesCatalog.put(fInfo)
            self.filesOperations.filesCatalog.sync()
            logs.debug("TMAN_SERVER: File " + filename + " succesfully acknowledged by peer " + str(peerId) + " )" + str(peerUsername) + ")")
        except Exception as e:
            logs.trace(e)
        

    def getTotalStorage(self):
        return self.perfParams.myTotalSpace

    def scoreLocations(self, locations, peerId, fileName):
        plac = ReplicaPlacement(ownerId = peerId, fileName = fileName, config = self.config)
        for loc in locations:
            try:
                peerPerfInfo = self.peersMonitor.getPerformanceInfo(loc)
            except KeyError:
                raise NotEnoughInformation()
            rl = RemoteLocation(serverHttpAddrArg = self.serverHttpAddr,
                                peerIdArg = loc,
                                avArg = peerPerfInfo.availability,
                                myAvArg = self.perfInfo.availability,
                                totalIncommingSize = peerPerfInfo.totalIncommingSize,
                                totalIncommingBwBound = peerPerfInfo.totalIncommingBwBound,
                                desiredBackWindow = self.perfParams.desiredBackWindow,
                                ttl = self.peersMonitor.getTtl(loc))
            plac.addPeer(rl)

        return plac.score()

    def scoreEachLocationExchange(self, locations, newLoc, newLocIncommingSize, fileName, alreadyRejected, totalIncommingSizes):
        #logs.debug("TMAN_SERVER: File to consider: " + str(fileName))
        allPlac = []

        plac = ReplicaPlacement(ownerId = self.myPeerId, fileName = fileName, config = self.config)
        if plac.resiliencyLevel > len(locations):
            return {}

        for loc in locations:
            try:
                peerPerfInfo = self.peersMonitor.getPerformanceInfo(loc)
            except KeyError:
                raise NotEnoughInformation()
            try:
                rejectedSize = alreadyRejected[loc]
            except KeyError:
                rejectedSize = 0
            if not totalIncommingSizes.has_key(loc):
                totalIncommingSizes[loc] = peerPerfInfo.totalIncommingSize
            assert(totalIncommingSizes[loc] != None)
            incommingSize = totalIncommingSizes[loc]

            #logs.debug("TMAN_SERVER:       Old: " + str(loc) + " incS: " + str(incommingSize - rejectedSize) + ", bw: " + str(peerPerfInfo.totalIncommingBwBound) + ", av: " + str(peerPerfInfo.availability) + ", ttl: " + str(self.peersMonitor.getTtl(loc)))

            rl = RemoteLocation(serverHttpAddrArg = self.serverHttpAddr,
                                peerIdArg = loc,
                                avArg = peerPerfInfo.availability,
                                myAvArg = self.perfInfo.availability,
                                totalIncommingSize = incommingSize - rejectedSize,
                                totalIncommingBwBound = peerPerfInfo.totalIncommingBwBound,
                                desiredBackWindow = self.perfParams.desiredBackWindow,
                                ttl = self.peersMonitor.getTtl(loc))
            plac.addPeer(rl)
            allPlac.append(rl)

        score = plac.score()
        newPeerPerfInfo = self.peersMonitor.getPerformanceInfo(newLoc)
        scoresImpr = {}

        for p in allPlac:
            plac.removePeer(p)

            try:
                rejectedSize = alreadyRejected[p.peerId]
            except KeyError:
                rejectedSize = 0

            newRl = RemoteLocation(serverHttpAddrArg = self.serverHttpAddr,
                                peerIdArg = newLoc,
                                avArg = newPeerPerfInfo.availability,
                                myAvArg = self.perfInfo.availability,
                                totalIncommingSize = newLocIncommingSize + rejectedSize,
                                totalIncommingBwBound = newPeerPerfInfo.totalIncommingBwBound,
                                desiredBackWindow = self.perfParams.desiredBackWindow,
                                ttl = self.peersMonitor.getTtl(newLoc))
            plac.addPeer(newRl)
            newScore = plac.score()
            plac.removePeer(newRl)
            plac.addPeer(p)

            #logs.debug("TMAN_SERVER:       New peer " + str(newLoc) + " incS: " + str(newLocIncommingSize + rejectedSize) + ", bw: " + str(newPeerPerfInfo.totalIncommingBwBound) + ", av: " + str(newPeerPerfInfo.availability) + ", ttl: " + str(self.peersMonitor.getTtl(newLoc)))
            #logs.debug("TMAN_SERVER:       For peer " + str(p.peerId) + " old score: " + str(score) + ", newScore: " + str(newScore) + ", score gain: " + str(newScore - score))
            scoresImpr[p.peerId] = newScore - score

        return scoresImpr


    def printAgreements(self):
        myProxy = xmlrpclib.ServerProxy("http://"+str(self.myAddr) + ":" + str(self.myPortNum), allow_none = True)
        agreementsSize = int(myProxy.getAgreementsSize())
        myPd = PeerDescriptor(self.serverHttpAddr, self.myPeerId)
        print(str(self.myPeerId) + " (" + str(myPd.username) + "):")
        print("    TotalIncommingSize: " + str(agreementsSize))
        print("    TotalIncommingBwBound: " + str(self.perfParams.totalIncommingBwBound) + ", ratio: " +\
              str(float(agreementsSize)/ (self.perfParams.totalIncommingBwBound * float(self.perfInfo.availability))))
        print("    Availability: " + str(self.perfInfo.availability))
        myProxy.printAgreements()
        myProxy.printDetailAgreements()
        print ""
        allFilesInfo = self.filesOperations.filesCatalog.getAllFileInfo()
        sizePerPeer = {}
        for v in allFilesInfo:
            print("    " + str(v.lastLocation) + "(" + str(v.size) + ") -> " + str(v.replicatorsIds))
            for r in v.replicatorsIds:
                if not sizePerPeer.has_key(r):
                    sizePerPeer[r] = 0
                sizePerPeer[r] += int(v.size)
        for k in sizePerPeer.keys():
            print("    " + str(k) + " -> " + str(sizePerPeer[k]))

