#! /usr/bin/env python

import getpass
import os
import logs
import random
import time
from os import path
from filesLocationsManager import FilesLocationsManager
from common import KeysManager
from common import get_size
from common import sizeScalingFactor
from negoClient import NegoClient
from peerDescriptor import PeerDescriptor
from filesMetadata import FileInfo, FilesCatalog, computeFileId
from asynchCommandClient import AsynchCommandClient
from utils import lock
import traceback
import xmlrpclib
import config

class CouldNotMakeBackupException(Exception):
    pass


class FilesOperations:
    def __init__(self, serverHttpAddrArg, metadataPath, homeDirArg, userPathArg, statsPathArg, myAddr, myPortNum, myPeerId):
        self.homeDir = homeDirArg
        self.userPath = userPathArg
        self.dhtAddress = serverHttpAddrArg
        self.keysManager = KeysManager(myAddr, self.homeDir, self.userPath)
        self.myPeerId = myPeerId
        self.negoClient = NegoClient(self.myPeerId, serverHttpAddrArg)
        self.statsPath = statsPathArg
        self.chunkSize = 10000
        self.username = getpass.getuser()
        self.myAddr = myAddr
        self.myPortNum = myPortNum

        self.filesLocationsManager = None
        self.basicPath = os.path.join(self.userPath, "peerstorage")
        self.metadataPath = metadataPath
        self.dataPath = os.path.join(self.basicPath, "data", "")
        self.configPath = os.path.join(self.basicPath, "config", "")
        self.catalogName = "filesCatalog-" + str(self.myPortNum)
        self.catalogFilename = os.path.join(self.metadataPath, self.catalogName)
        self.lockfile = lock(os.path.join(self.basicPath, "fileOperationsLock"))

        self.filesCatalog = FilesCatalog(self.catalogFilename, self.myAddr, self.myPortNum, self)

        if config.enviroment == "planetLab":
            self.sshOptions = "-o ConnectTimeout=2 -o 'StrictHostKeyChecking no' -o 'CheckHostIp no' -i planetLabKey.pub"
        else:
            self.sshOptions = "-o ConnectTimeout=2 -o 'StrictHostKeyChecking no' "

        #self.filesCatalog = None # backup will add FileInfo entries to this catalog unless it's None

    def getRealUsedStorage(self):
        return get_size(self.dataPath)

    def copyLocalFile(self, toUsername, toUsernameDir, toHostname, myPath):
        toDataDirectory = os.path.join(toUsernameDir, "peerstorage", "data", self.myPeerId)
        myPathDir = os.path.dirname(myPath)
        destinationPath = toDataDirectory + myPathDir
        createDirectoriesCommand = "ssh " + self.sshOptions + " %s@%s mkdir -p %s" % (toUsername, toHostname, destinationPath)
        copyCommand = "(scp -r " + self.sshOptions + " %s %s@%s:%s) 2> /dev/null > /dev/null" % (myPath, toUsername, toHostname, destinationPath)
        logs.debug("FILE_OPERATIONS: Executing command: " + str(createDirectoriesCommand))
        logs.debug("FILE_OPERATIONS: Executing command: " + str(copyCommand))
        ret1 = os.system(createDirectoriesCommand)
        ret2 = os.system(copyCommand)
        if (ret1 != 0) or (ret2 != 0):
            raise CouldNotMakeBackupException()

    def copyFileFrom(self, ownerId, fromUsername, fromHostname, fromUsernameDir, filename):
        fromDataDirectory = fromUsernameDir + "/peerstorage" + "/data/" + ownerId + "/"
        fromPath = fromDataDirectory + filename
        toFilename = self.dataPath + ownerId + filename
        toFilenameDir = os.path.dirname(toFilename)
        createDirectoriesCommand = "mkdir -p %s" % (toFilenameDir)
        copyCommand = "(scp -r " + self.sshOptions + " %s@%s:%s %s) 2> /dev/null > /dev/null" % (fromUsername, fromHostname, fromPath, toFilename)
        logs.debug("FILE_OPERATIONS: Executing command: " + str(createDirectoriesCommand))
        logs.debug("FILE_OPERATIONS: Executing command: " + str(copyCommand))
        ret1 = os.system(createDirectoriesCommand)
        ret2 = os.system(copyCommand)
        if (ret1 != 0) or (ret2 != 0):
            raise CouldNotMakeBackupException()
        #time.sleep(10)

    def downloadOriginalFileFrom(self, fromUsername, fromHostname, fromUsernameDir, filename):
        fromDataDirectory = fromUsernameDir + "/peerstorage" + "/data/" + self.myPeerId + "/"
        fromPath = fromDataDirectory + filename
        toFilename = filename
        toFilenameDir = os.path.dirname(toFilename)
        createDirectoriesCommand = "mkdir -p %s" % (toFilenameDir)
        copyCommand = "(scp -r " + self.sshOptions + " %s@%s:%s %s) 2> /dev/null > /dev/null" % (fromUsername, fromHostname, fromPath, toFilename)
        logs.debug("FILE_OPERATIONS: Executing command: " + str(createDirectoriesCommand))
        logs.debug("FILE_OPERATIONS: Executing command: " + str(copyCommand))
        ret1 = os.system(createDirectoriesCommand)
        ret2 = os.system(copyCommand)
        if (ret1 != 0) or (ret2 != 0):
            raise CouldNotMakeBackupException()

    def copyFileFromOwner(self, fromId, fromUsername, fromHostname, filename):
        toDataDir = self.dataPath + fromId + os.path.dirname(filename)
        toFilename = self.dataPath + fromId + filename
        createDirectoriesCommand = "mkdir -p %s" % (toDataDir)
        copyCommand = "(scp -r " + self.sshOptions + " %s@%s:%s %s) 2> /dev/null > /dev/null" % (fromUsername, fromHostname, filename, toFilename)
        logs.debug("FILE_OPERATIONS: Executing command: " + str(createDirectoriesCommand))
        logs.debug("FILE_OPERATIONS: Executing command: " + str(copyCommand))
        ret1 = os.system(createDirectoriesCommand)
        ret2 = os.system(copyCommand)
        if (ret1 != 0) or (ret2 != 0):
            raise CouldNotMakeBackupException()
        #time.sleep(10)

    def removeFile(self, ownerId, filename):
        diskLoc = os.path.join(self.dataPath, ownerId, filename)
        rmCommand = "(rm -rf %s) 2> /dev/null > /dev/null" % diskLoc
        logs.debug("Executing command: " + str(rmCommand))
        os.system(rmCommand)

    def copyReplicaFileFrom(self, fromUsername, fromUsernameDir, fromHostname, ownerUsername, ownerUsernameDir, myPath):
        fromDataDirectory = os.path.join(fromUsernameDir, "peerstorage", "data")
        toDataDirectory = os.path.join(ownerUsernameDir, "peerstorage", "data")
        myPathDir = os.path.dirname(myPath)
        fromPath = fromDataDirectory + myPath
        toPath = toDataDirectory+ myPathDir
        createDirectoriesCommand = "mkdir -p %s" % (toPath)
        copyCommand = "(scp -r " + self.sshOptions + " %s@%s:%s %s) 2> /dev/null > /dev/null" % (fromUsername, fromHostname, fromPath, toPath)
        logs.debug("FILE_OPERATIONS: Executing command: " + str(createDirectoriesCommand))
        logs.debug("FILE_OPERATIONS: Executing command: " + str(copyCommand))
        os.system(createDirectoriesCommand)
        os.system(copyCommand)


    def makeBackup(self, filesList, addStats):
        myNegoProxy = xmlrpclib.ServerProxy("http://"+str(self.myAddr) + ":" + str(self.myPortNum), allow_none = True)

        startBackupTime = time.time()
        for singleFile in filesList:
            self.backupSingleFile(singleFile, myNegoProxy)

        #if self.filesCatalog != None:
        #    self.filesCatalog.sync()
        finishBackupTime = time.time()
        backupDuration = finishBackupTime - startBackupTime
        if addStats:
            statsFile = open(self.statsPath, 'a')
            statsFile.write("backup-duration: " + str(backupDuration) + "\n")
            statsFile.close()


    def backupSingleFile(self, f, myNegoProxy):
        try:
            # just backuping at empty list - our backups are proactive
            size = os.path.getsize(f)
            version = 0
            fId = computeFileId(f)
            try: 
                fileInfo = self.filesCatalog.get(fId)
                version = fileInfo.version
            except KeyError:    
                logs.info("FILE_OPERATIONS: backuping file " + f) 
                fileInfo = FileInfo(f, self.myPeerId)
                fileInfo.version = version + 1
                fileInfo.size = str(sizeScalingFactor * os.path.getsize(f))
            myNegoProxy.setVersion(self.myPeerId, fileInfo.lastLocation, fileInfo.version)
            #self.updateSingleFile(f, fileInfo)
            if self.filesLocationsManager is not None:
                self.filesLocationsManager.updateScore(fileInfo)
            if self.filesCatalog != None:
                self.filesCatalog.put(fileInfo)
                self.filesCatalog.sync()
        except Exception as e:
            logs.trace(e)

    def changeLocationInFilesCatalog(self, fInfo, location, toRemove):
        if toRemove != None:
            i = fInfo.replicatorsIds.index(toRemove)
            fInfo.replicatorsIds[i] = location
        else:
            fInfo.replicatorsIds.append(location)
            while len(fInfo.committedLocation) < len(fInfo.replicatorsIds):
                fInfo.committedLocation.append((None, 0, True)) # third arg says if we already sent order for replication
            while len(fInfo.oldLocations) < len(fInfo.replicatorsIds):
                fInfo.oldLocations.append(None)
        self.filesCatalog.put(fInfo)
        self.filesCatalog.sync()

        #toSafeRemove = None
        #fInfo.replicatorsIds.append(location)
        #if toRemove is not None:
        #    fInfo.replicatorsIds.remove(toRemove)
        #if fInfo.oldLocations.has_key(toRemove):
            # This condition is to handle cycles (do not remove data from myself)
        #    if fInfo.oldLocations[toRemove] != location:
        #        fInfo.oldLocations[location] = fInfo.oldLocations[toRemove]
        #    toSafeRemove = toRemove
        #    del fInfo.oldLocations[toRemove]
        #else:
        #    fInfo.oldLocations[location] = toRemove
        #for oldId, toRemove in fInfo.oldLocations.items():
        #    if toRemove == location:
        #        del fInfo.oldLocations[oldId]
        #self.filesCatalog.put(fInfo)
        #self.filesCatalog.sync()
        #return toSafeRemove

    def changeFileLocation(self, fInfo, location, toRemove):
        remotePeer = PeerDescriptor(self.dhtAddress, location)
        self.copyLocalFile(remotePeer.username, remotePeer.dataDir, remotePeer.ipAddress, fInfo.lastLocation)
        fInfo.replicatorsIds.append(location)
        if toRemove is not None:
            removePeer = PeerDescriptor(self.dhtAddress, toRemove)
            self.removeFileFrom(removePeer.username, removePeer.dataDir, removePeer.ipAddress, fInfo.lastLocation)
            fInfo.replicatorsIds.remove(toRemove)
            negoProxy = xmlrpclib.ServerProxy("http://"+str(removePeer.ipAddress) + ":" + str(removePeer.portNumber), allow_none = True)
            negoProxy.rejectStorageContract(str(self.myPeerId), str(fInfo.size))
            
        self.filesCatalog.put(fInfo)
        self.filesCatalog.sync()

    # now this method assumes that all replicators are alive. TODO: if someone is dead take next replicator
    def restore(self, filesInfo):
        for f in filesInfo:
            assert len(f.replicatorsIds) > 0, "File lost: " + str(f)
            #TODO Do it for every replicator
            replicator = f.replicatorsIds[0]
            self.restoreFile(f.lastLocation, replicator)

    def restoreFileFrom(self, peerId, filename, version):
        if not os.path.exists(filename):
            peerDesc = PeerDescriptor(self.dhtAddress, peerId)
            self.downloadOriginalFileFrom(peerDesc.username, peerDesc.ipAddress, peerDesc.dataDir, filename)
        size = os.path.getsize(filename)
        fId = computeFileId(filename)
        try: 
            fileInfo = self.filesCatalog.get(fId)
        except KeyError:    
            fileInfo = FileInfo(filename, self.myPeerId)
            fileInfo.version = version
            fileInfo.size = str(sizeScalingFactor * os.path.getsize(filename))

        if self.filesCatalog != None:
            self.filesCatalog.put(fileInfo)
            self.filesCatalog.sync()

