#!/usr/bin/python
import logs
import xmlrpclib
import time
import os
import sys
import pickle
import socket
import getpass
import thread
import random
from os import path
from asynchCommand import FileVersionsCommand
from asynchCommandClient import AsynchCommandClient
from common import release, acquire
from filesMetadata import FileInfo
from filesMetadata import serializeFileInfo
from peerDescriptor import PeerDescriptor
from datetime import datetime
from SimpleXMLRPCServer import SimpleXMLRPCServer
from xmlrpclib import Fault
from common import KeysManager, calcResiliencyLevel
from persistentDict import PersistentDict

class NegoServer(object):
    """
    Passive (server) part of the negotation protocol
    """
    def __init__(self,
                 dhtAddr,
                 metadataPath,
                 myPeerId,
                 myUsername,
                 homeDirArg,
                 userPathArg,
                 statsPathArg,
                 myAddr,
                 port,
                 config,
                 filesOperations,
                 ndhtClient):
        self.metadataPath = metadataPath
        self.finishedThread = True
        self.username = myUsername
        self.myAddr = myAddr
        self.myPort = port
        self.aggreements = PersistentDict(metadataPath + "/agreements-" + str(self.myPort))
        self.aggreementsSize = 0
        for i in self.aggreements.keys([]):
            for j in self.aggreements.keys([i]):
                self.aggreementsSize += int(self.aggreements.get([i, j]))
        self.aggreementsSizeVerification = 0
        self.dhtAddress = dhtAddr
        self.homeDir = homeDirArg
        self.userPath = userPathArg
        self.statsPath = statsPathArg
        self.myPeerId = myPeerId
        self.keyManager = KeysManager(self.myAddr, self.homeDir, self.userPath)
        self.config = config
        self.filesOperations = filesOperations
        self.ndhtClient = ndhtClient
        self.dictLoc = thread.allocate_lock()
        self.oldAgreements = PersistentDict(metadataPath + "/oldAgreements-" + str(self.myPort))
        # this is a map: ownerId -> filename -> ([fromIds], version, deadline, onlyToConfirm)
        self.filesToReplicate = PersistentDict(metadataPath + "/filesToReplicate-" + str(self.myPort))
        # this is a map: ownerId -> filename -> version
        self.filesVersions = PersistentDict(metadataPath + "/filesVersions-" + str(self.myPort))
        # this is a map: fromId -> [(filename, deadline)]
        self.removeFilesCommand = PersistentDict(metadataPath + "/removeFiles-" + str(self.myPort))
        self.asynchCommandClient = AsynchCommandClient(self.dhtAddress,
                                                       self.myPeerId,
                                                       self.myAddr,
                                                       self.myPort,
                                                       self.statsPath,
                                                       self.config)
        self.lastInconsistencyCheck = {} # this is a map: ownerId -> timestamp
        self.lastInconsistencyOutCheck = {} # this is a map: ownerId -> timestamp
        socket.setdefaulttimeout(self.config.socketTimeout)

    def isWorking(self):
        return True

    def rejectStorageContract(self, remoteId, filesList):
        try:
            for f in filesList:
                try:
                    s = self.aggreements.get([remoteId, f])
                    self.aggreements.remove([remoteId, f])
                    logs.debug("NEGO-SERVER: Removed storage agreement for file " + f)
                    self.aggreementsSize -= int(s)
                except: 
                    continue
                if self.filesVersions.has_key([remoteId, f]) or self.filesToReplicate.has_key([remoteId, f]):
                    self.oldAgreements.set([remoteId, f], str(s))
            self.aggreements.sync()
            self.oldAgreements.sync()

            return 1
        except Exception as e:
            logs.trace(e)

    def acceptStorageContract(self, remoteId, size, filesList):
        remotePeer = PeerDescriptor(self.dhtAddress, remoteId)
        
        for e in filesList:
            f, s = e
            if not self.aggreements.has_key([remoteId, f]):
                self.aggreementsSize += int(s) 
                self.aggreements.set([remoteId, f], str(s))
                logs.debug("NEGO-SERVER: Added storage agreement for file " + f)
                #TODO: now the files does not change their sizes
            try:
                self.oldAgreements.remove([remoteId, f])
            except:
                pass
            remotePeer = PeerDescriptor(self.dhtAddress, remoteId)
        self.aggreements.sync()
        self.oldAgreements.sync()
        self.keyManager.authorizeKey(remotePeer.publicKey)
        logs.debug("NEGO-SERVER: added public key for peer: "+str(remoteId)+" key: "+str(remotePeer.publicKey))

    def getReplicators(self, ownerId, filename):
        try:
            acquire(self.dictLoc)
            if not self.filesToDownload.has_key(ownerId):
                release(self.dictLoc)
                return []
            if not self.filesToDownload[ownerId].has_key(filename):
                release(self.dictLoc)
                return []
            locs, deadline = self.filesToDownload[ownerId][filename]
            self.filesToDownload[ownerId][filename] = (locs, 10000000000000000)
            self.filesToDownload.sync()
            release(self.dictLoc)
            return locs
        except Exception as e:
            release(self.dictLoc)
            raise e

    def repairInconsistentContracts(self):
        thread.start_new_thread(self.repairInconsistentContracts_thread, ())

    def repairInconsistentContracts_thread(self):
        self.repairIncommingContracts()
        self.repairOutcommingContracts()
        self.cleanUpDanglingData()

    def repairIncommingContracts(self):
        try:
            infoPerOwner = {}
            for o in self.filesVersions.keys([]):
                if o != self.myPeerId:
                    if not infoPerOwner.has_key(o):
                        infoPerOwner[o] = {'fileVersions': {}, 'oldAgreements': {}, 'agreements': {}}
                    for f in self.filesVersions.keys([o]):
                        infoPerOwner[o]['fileVersions'][f] = self.filesVersions.get([o, f])

            for o in self.aggreements.keys([]):
                if o != self.myPeerId:
                    if not infoPerOwner.has_key(o):
                        infoPerOwner[o] = {'fileVersions': {}, 'oldAgreements': {}, 'agreements': {}}
                    for f in self.aggreements.keys([o]):
                        infoPerOwner[o]['agreements'][f] = str(self.aggreements.get([o, f]))
        
            for o in self.oldAgreements.keys([]):
                if o != self.myPeerId:
                    if not infoPerOwner.has_key(o):
                        infoPerOwner[o] = {'fileVersions': {}, 'oldAgreements': {}, 'agreements': {}}
                    for f in self.oldAgreements.keys([o]):
                        infoPerOwner[o]['oldAgreements'][f] = str(self.oldAgreements.get([o, f]))

            current = time.time()
            for o in infoPerOwner.keys():
                try:
                    lastCheck = self.lastInconsistencyCheck[o]
                except KeyError:
                    lastCheck = 0
                if (current - lastCheck > self.config.contractsInconsistencyCheckIntervalPerOwnerSec):
                    try:
                        remotePeer = PeerDescriptor(self.dhtAddress, str(o))
                        self.ndhtClient.updateWithPeer(remotePeer.ipAddress, int(remotePeer.portNumber))
                        remoteProxy = xmlrpclib.ServerProxy("http://"+str(remotePeer.ipAddress) + ":" + str(remotePeer.portNumber), allow_none = True)
                        remoteProxy2 = xmlrpclib.ServerProxy("http://"+str(remotePeer.ipAddress) + ":" + str(int(remotePeer.portNumber) + 2),
                                                             allow_none = True)
                        dhtProxy = xmlrpclib.ServerProxy(self.dhtAddress, allow_none=True)
                        self.filesOperations.filesLocationsManager.preparePerfromanceInfo()
                        ver = dhtProxy.getVersion(self.myPeerId)
                        remoteProxy2.updatePerformanceInfo(self.filesOperations.filesLocationsManager.perfInfo, ver)
                        remoteProxy.verifyOutcommingContractsFrom(
                            self.myPeerId,
                            infoPerOwner[o]['fileVersions'],
                            infoPerOwner[o]['agreements'],
                            infoPerOwner[o]['oldAgreements']
                        )
                        self.lastInconsistencyCheck[o] = time.time()
                    except socket.error as e:
                        logs.commonErr(e)
                    except Exception as e:
                        logs.trace(e)
        except Exception as e:
            logs.trace(e)

    def repairOutcommingContracts(self):
        try:
            infoPerOwner = {}
            allFilesInfo = self.filesOperations.filesCatalog.getAllFileInfo()
            for fInfo in allFilesInfo:
                locs = fInfo.committedLocation
                for l in locs:
                    peerId = l[0]
                    if (peerId != None) and (peerId != self.myPeerId):
                        if not infoPerOwner.has_key(peerId):
                            infoPerOwner[peerId] = {'fileVersions': {}, 'agreements': {}, 'committed': {}}
                        infoPerOwner[peerId]['committed'][fInfo.lastLocation] = fInfo.size
                for peerId in fInfo.replicatorsIds:
                    if (peerId != None) and (peerId != self.myPeerId):
                        if not infoPerOwner.has_key(peerId):
                            infoPerOwner[peerId] = {'fileVersions': {}, 'agreements': {}, 'committed': {}}
                        infoPerOwner[peerId]['agreements'][fInfo.lastLocation] = fInfo.size
                for peerId in fInfo.oldLocations:
                    if (peerId != None) and (peerId != self.myPeerId):
                        if not infoPerOwner.has_key(peerId):
                            infoPerOwner[peerId] = {'fileVersions': {}, 'agreements': {}, 'committed': {}}
                        infoPerOwner[peerId]['fileVersions'][fInfo.lastLocation] = fInfo.size

            current = time.time()
            for o in infoPerOwner.keys():
                try:
                    lastCheck = self.lastInconsistencyOutCheck[o]
                except KeyError:
                    lastCheck = 0
                if (current - lastCheck > self.config.contractsInconsistencyCheckIntervalPerOwnerSec):
                    try:
                        remotePeer = PeerDescriptor(self.dhtAddress, str(o))
                        self.ndhtClient.updateWithPeer(remotePeer.ipAddress, int(remotePeer.portNumber))
                        remoteProxy = xmlrpclib.ServerProxy("http://"+str(remotePeer.ipAddress) + ":" + str(remotePeer.portNumber), allow_none = True)
                        remoteProxy2 = xmlrpclib.ServerProxy("http://"+str(remotePeer.ipAddress) + ":" + str(int(remotePeer.portNumber) + 2),
                                                             allow_none = True)
                        dhtProxy = xmlrpclib.ServerProxy(self.dhtAddress, allow_none=True)
                        self.filesOperations.filesLocationsManager.preparePerfromanceInfo()
                        ver = dhtProxy.getVersion(self.myPeerId)
                        remoteProxy2.updatePerformanceInfo(self.filesOperations.filesLocationsManager.perfInfo, ver)
                        remoteProxy.verifyIncommingContractsFrom(
                            self.myPeerId,
                            infoPerOwner[o]['fileVersions'],
                            infoPerOwner[o]['agreements'],
                            infoPerOwner[o]['committed']
                        )
                        self.lastInconsistencyOutCheck[o] = time.time()
                    except socket.error as e:
                        logs.commonErr(e)
                    except Exception as e:
                        logs.trace(e)
        except Exception as e:
            logs.trace(e)

    def cleanUpDanglingData(self):
        #TODO remove from filesystems these files that are not in self.filesVersions evidence
        pass

    def verifyContracts(self):
        thread.start_new_thread(self.verifyContracts_thread, ())

    def verifyContracts_thread(self):
        allFilesInfo = self.filesOperations.filesCatalog.getAllFileInfo()
        filesByReplicators = {}
        for fInfo in allFilesInfo:
            locs = fInfo.committedLocation
            replicators = locs + []
            replicators.append(self.myPeerId)
            for l in locs:
                if not filesByReplicators.has_key(l):
                    filesByReplicators[l] = []
                replicators.remove(l)
                filesByReplicators[l].append(serializeFileInfo(fInfo))
                replicators.append(l)
        for peerId in filesByReplicators:
            try:
                self.removeFromToRemove(peerId, filesByReplicators[peerId])
                remotePeer = PeerDescriptor(self.dhtAddress, str(peerId))
                remoteProxy = xmlrpclib.ServerProxy("http://"+str(remotePeer.ipAddress) + ":" + str(remotePeer.portNumber), allow_none = True)
                remoteProxy.verifyFileVersions(self.myPeerId, filesByReplicators[peerId])
            except Exception as e:
                logs.trace(e)
                try:
                    key = time.time()
                    asyncCommand = FileVersionsCommand(key)
                    asyncCommand.setOwner(self.myPeerId)
                    asyncCommand.setFileVersions(filesByReplicators[peerId])
                    res = self.asynchCommandClient.sendCommandForPeer(peerId, "FileVersionsCommand", asyncCommand)
                    if res:
                        logs.info("NEGO-SERVER: sent command for file version verification to peer: " + str(peerId))
                    else:
                        logs.info("NEGO-SERVER: unable to initialize asynch command to peer: " + str(peerId))
                except Exception as e:
                    logs.trace(e)

    def wakeup(self):
        if self.finishedThread:
            self.finishedThread = False
            thread.start_new_thread(self.wakeup_thread, ())

    def wakeup_thread(self):
        try:
            logs.info("NEGO-SERVER: wakeup - removing files")
            current = time.time()

            acquire(self.dictLoc)
            fileIds = self.removeFilesCommand.keys([])
            release(self.dictLoc)
            byDeadlines = []
            for fromId in fileIds:
                assert fromId != self.myPeerId
                acquire(self.dictLoc)
                entries = self.removeFilesCommand.get([fromId]) + []
                release(self.dictLoc)
                for p in entries:
                    filename, deadline = p
                    if deadline < current:
                        byDeadlines.append((deadline, filename, fromId))
            byDeadlines.sort()
            for e in byDeadlines:
                deadline, filename, fromId = e
                try:
                    remotePeer = PeerDescriptor(self.dhtAddress, str(fromId))
                    remoteProxy = xmlrpclib.ServerProxy("http://"+str(remotePeer.ipAddress) + ":" + str(remotePeer.portNumber), allow_none = True)
                    remoteProxy.removeFile(self.myPeerId, filename)
                    acquire(self.dictLoc)
                    l = self.removeFilesCommand.get([fromId])
                    try:
                        l.remove(p)
                    except:
                        pass
                    self.removeFilesCommand.set([fromId], l)
                except socket.error as e:
                    logs.commonErr(e)
                    l = self.removeFilesCommand.get([fromId])
                    try:
                        l.remove(p)
                    except:
                        pass
                    l.append((filename, deadline + self.config.intervalBeetweenNegoServerAttemptsSec))
                    self.removeFilesCommand.set([fromId], l)
                except Exception as e:
                    logs.trace(e)
                    l = self.removeFilesCommand.get([fromId])
                    try:
                        l.remove(p)
                    except:
                        pass
                    l.append((filename, deadline + self.config.intervalBeetweenNegoServerAttemptsSec))
                    self.removeFilesCommand.set([fromId], l)
                release(self.dictLoc)
            self.removeFilesCommand.sync()

            logs.info("NEGO-SERVER: wakeup - replicating files")

            acquire(self.dictLoc)
            owners = self.filesToReplicate.keys([])
            random.shuffle(owners)
            release(self.dictLoc)
            for ownerId in owners:
                acquire(self.dictLoc)
                files = self.filesToReplicate.keys([ownerId])
                release(self.dictLoc)
                for filename in files:
                    acquire(self.dictLoc)
                    try:
                        locs, version, deadline, forConfirmation = self.filesToReplicate.get([ownerId, filename])
                    except:
                        release(self.dictLoc)
                        continue
                    release(self.dictLoc)
                    if deadline < current:
                        if forConfirmation:
                            try:
                                ownerPeer = PeerDescriptor(self.dhtAddress, str(ownerId))
                                ownerProxy = xmlrpclib.ServerProxy(
                                                  "http://"+str(ownerPeer.ipAddress) + ":" + str(ownerPeer.portNumber + 2),
                                                  allow_none = True
                                             )
                                ownerProxy.ackDownload(filename, self.myPeerId, self.username)
                                logs.debug("NEGO_SERVER: Confirmed replication of file: " + filename)
                                acquire(self.dictLoc)
                                if (not self.filesVersions.has_key([ownerId, filename])) or (version > self.filesVersions.get([ownerId, filename])):
                                    self.filesToReplicate.set([ownerId, filename], (locs, version, deadline, False))
                                    logs.debug("NEGO_SERVER: The file version of file " + filename + " was not correct")
                                else:
                                    self.filesToReplicate.remove([ownerId, filename])
                                release(self.dictLoc)
                                self.filesToReplicate.sync()
                            except socket.error as e:
                                release(self.dictLoc)
                                logs.debug("NEGO_SERVER: Could not acknowledge download of the file; error " + str(e))
                                logs.commonErr(e)
                            except Exception as e:
                                release(self.dictLoc)
                                logs.debug("NEGO_SERVER: Could not acknowledge download of the file; error " + str(e))
                        else:
                            succ = False
                            logs.debug("NEGO_SERVER: Trying to download file " + filename + " from locations: " + str(locs))
                            for fromPeerId in locs:
                                try:
                                    remotePeer = PeerDescriptor(self.dhtAddress, str(fromPeerId))
                                    remoteProxy = xmlrpclib.ServerProxy(
                                                      "http://"+str(remotePeer.ipAddress) + ":" + str(remotePeer.portNumber),
                                                      allow_none = True
                                                  )
                                    remoteVer = remoteProxy.getVersion(ownerId, filename)
                                    try:
                                        ver = self.filesVersions.get([ownerId, filename])
                                    except KeyError:
                                        ver = -1
                                    if (remoteVer >= ver):
                                        logs.debug("NEGO_SERVER: The version for file " + filename + " is ok")
                                        succ = self.downloadFileLocal(ownerId, filename, fromPeerId)
                                        if succ and remoteVer >= version:
                                            acquire(self.dictLoc)
                                            self.filesToReplicate.set([ownerId, filename], (locs, version, deadline, True))
                                            release(self.dictLoc)
                                            self.filesToReplicate.sync()
                                            logs.debug("NEGO_SERVER: File " + filename + " downloaded correctly")
                                            break
                                    else:
                                        logs.debug("NEGO_SERVER: The version of file " + filename + ", and location: " + fromPeerId +\
                                                   " is incorrect; got " + str(remoteVer) + ", expected: " + str(version))
                                except socket.error as e:
                                    logs.commonErr(e)
                                except Exception as e:
                                    logs.trace(e)
                            if not succ:
                                acquire(self.dictLoc)
                                self.filesToReplicate.set([ownerId, filename],
                                                          (locs,
                                                           version,
                                                           time.time() + self.config.intervalBeetweenNegoServerAttemptsSec,
                                                           False))
                                release(self.dictLoc)
                                self.filesToReplicate.sync()

        except Exception as e:
            release(self.dictLoc)
            self.finishedThread = True
            logs.trace(e)
        self.finishedThread = True

    def removeFileLocal(self, ownerId, filename):
        self.filesOperations.removeFile(ownerId, filename)
        try:
            self.filesVersions.remove([ownerId, filename])
            self.filesVersions.sync()
        except:
            pass
        logs.debug("NEGO-SERVER: File: " + filename + " removed permanently")

    def getVersion(self, ownerId, filename):
        try:
            res = self.filesVersions.get([ownerId, filename])
            return res
        except:
            return -1

    def getMyFilesVersions(self):
        try:
            vers = {}
            if not self.filesVersions.has_key([self.myPeerId]):
                return vers
            for f in self.filesVersions.keys([self.myPeerId]):
                if not "filesCatalog" in path.basename(f):
                    vers[f] = self.filesVersions.get([self.myPeerId, f])
            return vers
        except Exception as e:
            logs.trace(e)
            raise e

    def getOthersFilesVersions(self):
        vers = {}
        for ownerId in self.filesVersions.keys([]):
            if ownerId != self.myPeerId:
                if not vers.has_key(ownerId):
                    vers[ownerId] = {}
                try:
                    for f in self.filesVersions.keys([ownerId]):
                        if not "filesCatalog" in path.basename(f):
                            try:
                                vers[ownerId][f] = self.filesVersions.get([ownerId, f])
                            except Exception as e:
                                logs.trace(e)
                except Exception as e:
                    logs.trace(e)
        return vers

    def setVersion(self, ownerId, filename, version):
        self.filesVersions.set([ownerId, filename], version)
        self.filesVersions.sync()
        #logs.debug("NEGO_SERVER: Set version for owner: " + str(ownerId) + ", filename: " + filename + ", version: " + str(version))

    def downloadFileLocal(self, ownerId, filename, fromId):
        try:
            remotePeer = PeerDescriptor(self.dhtAddress, str(fromId))

            if fromId == ownerId:
                self.filesOperations.copyFileFromOwner(fromId, remotePeer.username, remotePeer.ipAddress, filename)
            else:
                self.filesOperations.copyFileFrom(ownerId, remotePeer.username, remotePeer.ipAddress, remotePeer.dataDir, filename)

            remoteProxy = xmlrpclib.ServerProxy("http://"+str(remotePeer.ipAddress) + ":" + str(remotePeer.portNumber), allow_none = True)
            self.filesVersions.set([ownerId, filename], remoteProxy.getVersion(ownerId, filename))
            self.filesVersions.sync()
            return True
        except socket.error:
            return False

    def removeFile(self, ownerId, filename):
        try:
            acquire(self.dictLoc)
            try:
                self.filesToReplicate.remove([ownerId, filename])
            except KeyError:
                pass
            release(self.dictLoc)
            self.filesToReplicate.sync()       
            self.removeFileLocal(ownerId, filename)
            try:
                self.oldAgreements.remove([ownerId, filename])
            except:
                pass # the peer could be rejected than accepted and after accept it could receive removeRequest
            try:
                self.aggreements.remove([ownerId, filename])
                logs.debug("NEGO-SERVER: Removed storage agreement for file " + filename)
            except:
                pass # the peer could be rejected than accepted and after accept it could receive removeRequest
            self.aggreements.sync()
            self.oldAgreements.sync()
            logs.debug("NEGO-SERVER: File: " + filename + " from peer " + str(ownerId) + " completely removed")
        except Exception as e:
            logs.trace(e)
            raise e

    def printDicts(self, f):
        f.write(str(self.myPeerId) + " file versions:\n")
        for o in self.filesVersions.keys([]):
            f.write("    " + str(o) + "\n")
            for filename in self.filesVersions.keys([o]):
                f.write("         " + str(filename) + ", version: " + str(self.filesVersions.get([o, filename])) + "\n")
        f.write(str(self.myPeerId) + " files to replicate:\n")
        for o in self.filesToReplicate.keys([]):
            f.write("    " + str(o) + "\n")
            for filename in self.filesToReplicate.keys([o]):
                f.write("         " + str(filename) + ", locations: " + str(self.filesToReplicate.get([o, filename])[0]) + "\n")


    def getInfoFromReports(self):
        try:
            reportFilename = self.metadataPath + '/report-' + str(self.myPort) + '.txt'
            f = open(reportFilename, 'r')
            succ = 0
            lost = 0
            wait = 0
            for line in f:
                if "This file here is lost" in line:
                    lost += 1
                if "verified successfully" in line:
                    succ += 1
                if "It will be replicated" in line:
                    wait += 1
            return(succ, wait, lost)
        except:
            return (0, 0, 0)


    def verifyIncommingContractsFrom(self, peerId, fileVers, agr, commi):
        try:
            logs.debug("Received in verification request from peer " + str(peerId))
            remotePeer = PeerDescriptor(self.dhtAddress, str(peerId)) 
            remoteProxy = xmlrpclib.ServerProxy("http://"+str(remotePeer.ipAddress) + ":" + str(remotePeer.portNumber), allow_none = True)
            agreementsToDrop = []
            for f in agr.keys():
                if not self.aggreements.has_key([peerId, f]):
                    agreementsToDrop.append(f)
            remoteProxy.dropContract(self.myPeerId, agreementsToDrop)
        except Exception as e:
            logs.trace(e)


    def verifyOutcommingContractsFrom(self, peerId, fileVers, agr, oldAgr):
        try:
            logs.debug("Received out verification request from peer " + str(peerId))
            for f in fileVers.keys():
                logs.debug("Considering file from file versions: " + str(f))
                try:
                    fileInfo = self.filesOperations.filesCatalog.find(f)
                except IndexError:
                    logs.debug("NEGO_SERVER: Restoring file: " + str(f) + " from peer: " + str(peerId))
                    self.filesOperations.restoreFileFrom(peerId, f, fileVers[f])
                    self.filesVersions.set([self.myPeerId, f], fileVers[f])
                    fileInfo = self.filesOperations.filesCatalog.find(f)
                if oldAgr.has_key(f) or agr.has_key(f):
                    if (peerId not in fileInfo.oldLocations) and (len(fileInfo.oldLocations) < calcResiliencyLevel(f)):
                        fileInfo.oldLocations.append(peerId)
                    while len(fileInfo.committedLocation) < len(fileInfo.oldLocations):
                        try:
                            peerToAdd = fileInfo.oldLocations[len(fileInfo.committedLocation)]
                            ind = -1
                            for l in fileInfo.committedLocation:
                                if l[0] == peerToAdd:
                                    ind = fileInfo.committedLocation.index(l)
                            if ind == -1:
                                fileInfo.committedLocation.append((peerToAdd, time.time(), True))
                            else:
                                break
                        except Exception as e:
                            logs.trace(e)
                            
                    #while len(fileInfo.replicatorsIds) < len(fileInfo.committedLocation):
                    #    fileInfo.replicatorsIds.append(fileInfo.committedLocation[len(fileInfo.replicatorsIds)][0])
                self.filesOperations.filesCatalog.put(fileInfo)
            self.filesOperations.filesCatalog.sync()

            for f in oldAgr.keys():
                logs.debug("Considering file from old agreements: " + str(f))
                try:
                    fileInfo = self.filesOperations.filesCatalog.find(f)
                    ind = -1
                    for l in fileInfo.committedLocation:
                        if l[0] == peerId:
                            ind = fileInfo.committedLocation.index(l)
                    if (peerId not in fileInfo.oldLocations) and (ind == -1):
                        self.addToRemove(peerId, self.username, f)
                        logs.debug("NEGO_SERVER: Repairing inconsistent contract (cause 1) removing file: " + str(f) + " from peer: " + str(peerId))
                except IndexError:
                    continue

            agreementsToDrop = []
            for f in agr.keys():
                logs.debug("Considering file from current agreements: " + str(f))
                try:
                    fileInfo = self.filesOperations.filesCatalog.find(f)
                except IndexError:
                    agreementsToDrop.append(f)
                    logs.debug("NEGO_SERVER: Repairing inconsistent contract (cause 2) removing file: " + str(f) + " from peer: " + str(peerId))
                if peerId not in fileInfo.replicatorsIds:
                    agreementsToDrop.append(f)
                    logs.debug("NEGO_SERVER: Repairing inconsistent contract (cause 3) removing file: " + str(f) + " from peer: " + str(peerId))
            remotePeer = PeerDescriptor(self.dhtAddress, str(peerId)) 
            remoteProxy = xmlrpclib.ServerProxy("http://"+str(remotePeer.ipAddress) + ":" + str(remotePeer.portNumber), allow_none = True)
            remoteProxy.rejectStorageContract(self.myPeerId, agreementsToDrop)

        except Exception as e:
            logs.trace(e)

    def dropContract(self, peerId, agreementsToDrop):
        try:
            logs.debug("NEGO_SERVER: Contracts dropped for files: " + str(agreementsToDrop) + " from peer " + str(peerId))
            for f in agreementsToDrop:
                try:
                    fileInfo = self.filesOperations.filesCatalog.find(f)
                    if peerId in fileInfo.replicatorsIds:
                        try:
                            fileInfo.replicatorsIds.remove(peerId)
                        except:
                            pass
                    self.filesOperations.filesCatalog.put(fileInfo)
                except IndexError:
                    continue
            self.filesOperations.filesCatalog.sync()
        except Exception as e:
            logs.trace(e)

    def verifyFileVersions(self, ownerId, fileInfoList):
        try:
            totalSize = 0.0
            for fInfoDict in fileInfoList:
                fInfo = FileInfo(fInfoDict['lastLocation'], fInfoDict['ownerId'])
                fInfo.initFromDict(fInfoDict)
                totalSize += fInfo.size
                filename = fInfo.lastLocation
                version = fInfo.version
                try:
                    ver = self.filesVersions.get([ownerId, filename])
                    if ver < version:
                        self.replicateFilesFrom(ownerId, [fInfoDict]) # and here is update mechanism :)
                    continue
                except KeyError as e:
                    pass
            f.close()
        except Exception as e:
            logs.trace(e)
            raise e

    def rejectStorageContractIf(self, remoteId, size, filesList, expectedIncommingSize):
        currentIncommingSize = float(self.getAgreementsSize())
        if currentIncommingSize < (1.0 - self.config.changeStorageContractTollerance) * float(expectedIncommingSize):
            return 0
        else:
            for e in filesList:
                f, s = e
                try:
                    self.aggreements.remove([remoteId, f])
                    logs.debug("NEGO-SERVER: Removed storage agreement for file " + f)
                    self.aggreementsSize -= int(s)
                except: 
                    pass
                if self.filesVersions.has_key([remoteId, f]) or self.filesToReplicate.has_key([remoteId, f]):
                    self.oldAgreements.set([remoteId, f], str(s))
            self.aggreements.sync()
            self.oldAgreements.sync()

            return 1

    def acceptStorageContractIf(self, remoteId, size, expectedIncommingSize, filesList):
        currentIncommingSize = int(self.getAgreementsSize())
        if currentIncommingSize > (1.0 + self.config.changeStorageContractTollerance) * float(expectedIncommingSize):
            return 0
        else:
            
            for e in filesList:
                f, s = e
                if not self.aggreements.has_key([remoteId, f]):
                    self.aggreementsSize += int(s) 
                    self.aggreements.set([remoteId, f], str(s))
                    logs.debug("NEGO-SERVER: Added storage agreement for file " + f)
                    #TODO: now the files does not change their sizes
                try:
                    self.oldAgreements.remove([remoteId, f])
                except:
                    pass
            self.aggreements.sync()
            self.oldAgreements.sync()
            remotePeer = PeerDescriptor(self.dhtAddress, remoteId)
            self.keyManager.authorizeKey(remotePeer.publicKey)
            return 1

    def replicateFilesFrom(self, ownerId, fileInfoList):
        try:
            acquire(self.dictLoc)
            deadline = time.time() + self.config.downloadFilesHangTimeSec
            for fInfoDict in fileInfoList:
                fInfo = FileInfo(fInfoDict['lastLocation'], fInfoDict['ownerId'])
                fInfo.initFromDict(fInfoDict)
                try:
                    my_ver = self.filesVersions.get([ownerId, fInfo.lastLocation])
                except KeyError:
                    self.filesVersions.set([ownerId, fInfo.lastLocation], -1)
                    my_ver = -1
                if my_ver >= fInfo.version:
                    logs.debug("NEGO-SERVER: " + fInfo.lastLocation + " is already here")
                    continue

                reps = []
                reps.append(ownerId)
                i = 0
                try:
                    for l in fInfo.oldLocations:
                        if l != None:
                            reps.append(l)
                            reps.append(fInfo.committedLocation[i])
                        i += 1
                except IndexError:
                    pass
                    
                if self.myPeerId in reps:
                    reps.remove(self.myPeerId)

                if self.filesToReplicate.has_key([ownerId, fInfo.lastLocation]):
                    r1, ver1, d1, toConf =  self.filesToReplicate.get([ownerId, fInfo.lastLocation])
                    if fInfo.version > ver1:
                        self.filesToReplicate.set([ownerId, fInfo.lastLocation], (reps, fInfo.version, min(d1, deadline), False))
                        logs.debug("NEGO-SERVER: " + fInfo.lastLocation + " <- added for replication with deadline: " +\
                                   str(datetime.fromtimestamp(deadline)) + " and version: " + str(fInfo.version))
                    else:
                        self.filesToReplicate.set([ownerId, fInfo.lastLocation], (reps, ver1, d1, toConf))
                else:
                    self.filesToReplicate.set([ownerId, fInfo.lastLocation], (reps, fInfo.version, deadline, False))
                    logs.debug("NEGO-SERVER: " + fInfo.lastLocation + " <- added for replication with deadline: " +\
                               str(datetime.fromtimestamp(deadline)) + " and version: " + str(fInfo.version))

            release(self.dictLoc)
            self.filesToReplicate.sync()
        except Exception as e:
            release(self.dictLoc)
            logs.trace(e)
            raise e

    def addToRemove(self, peerId, peerUsername, filename):
        acquire(self.dictLoc)
        if self.removeFilesCommand.has_key([peerId]):
            l = self.removeFilesCommand.get([peerId])
        else:
            l = []
        add = True
        for e in l:
            f, d = e
            if f == filename:
                add = False
                break
        if add:
            l.append((filename, time.time()))
            self.removeFilesCommand.set([peerId], l)
        release(self.dictLoc)
        self.removeFilesCommand.sync()
        logs.debug("NEGO-SERVER: File: " + filename + " from peer " + str(peerId) + " (" + peerUsername + ") added to be removed")

    def removeFromToRemove(self, peerId, filenames):
        for filename in filenames:
            acquire(self.dictLoc)
            if not self.removeFilesCommand.has_key([peerId]):
                release(self.dictLoc)
                return
            l = self.removeFilesCommand.get([peerId])
            for e in l:
                f, d = e
                if f == filename:
                    l.remove(e)
            self.removeFilesCommand.set([peerId], l)
            release(self.dictLoc)
        self.removeFilesCommand.sync()
        logs.debug("NEGO-SERVER: Files: " + str(filenames) + " from peer " + str(peerId) + " removed from to-remove-list")


    def authenticateForRestore(self, peerId):
        remotePeer = PeerDescriptor(self.dhtAddress, peerId)
        self.keyManager.authorizeKey(remotePeer.publicKey)

    def getOldAgreementsSize(self):
        size = 0
        for ownerId in self.oldAgreements.keys([]):
            for f in self.oldAgreements.keys([ownerId]):
                try:
                    size += int(self.oldAgreements.get([ownerId, f]))
                except:
                    pass
        return str(size)

    def getAgreementsSize(self):
        self.aggreementsSizeVerification += 1
        if self.aggreementsSizeVerification == 30:
            res = 0
            for i in self.aggreements.keys([]):
                for j in self.aggreements.keys([i]):
                    res += int(self.aggreements.get([i, j]))
            self.aggreementsSize = res
            self.aggreementsSizeVerification = 0
        return str(self.aggreementsSize)

    def printAgreements(self):
        for k in self.aggreements.keys([]):
            sizes = self.aggreements.get([k])
            s = 0
            for e in sizes.values():
                s += int(e)
            print("    " + str(k) + " <- " + str(s))

    def printDetailAgreements(self):
        for k in self.aggreements.keys([]):
            sizes = self.aggreements.get([k])
            s = 0
            print("    " + str(k) + ":")
            for e in sizes.keys():
                print("        " + str(e) + " <- " + str(sizes[e]))

    def printFileVersions(self):
        print "File versions for peer: " + str(self.myPeerId)
        for ownerId in self.filesVersions.keys([]):
            print "    Owner: " + str(ownerId)
            for filename in self.filesVersions.keys([ownerId]):
                print "        " + filename + " <- " + str(self.filesVersions.get([ownerId, filename]))

    def registerInDht(self, myPeerId = None):

        if myPeerId == None:
            myPeerId = self.myPeerId
        if myPeerId == None:
            raise ValueError("myPeerID not defined")

        logs.info("NEGO-SERVER: Registering in dht port: " + str(self.port) + ", ipAddress: " + str(self.ipAddress) + ", username: " + str(self.username))
        dhtProxy = xmlrpclib.ServerProxy(self.dhtAddress, allow_none=True)
        dhtProxy.putField(str(myPeerId), "portNumber", self.port)
        dhtProxy.putField(str(myPeerId), "ipAddress", self.ipAddress)
        dhtProxy.putField(str(myPeerId), "username", self.username)


def startNegoServer(dhtAddr,
                    metadataPath,
                    myPeerId,
                    myUsername,
                    ipAddr,
                    port,
                    homeDir,
                    userPath,
                    statsPath,
                    config,
                    filesOperations,
                    ndhtClient,
                    totalStorage = None,
                    dhtAddress = None):
    negoServer = NegoServer(dhtAddr, metadataPath, myPeerId, myUsername, homeDir, userPath, statsPath, ipAddr, port, config, filesOperations, ndhtClient)
    if (totalStorage == None):
        totalStorage = 25000 * 2**20
    negoServer.totalStorage = totalStorage

    negoServer.dhtAddress = dhtAddress
    negoServer.port = port
    negoServer.ipAddress = ipAddr
    negoServer.username = getpass.getuser()
    
    rpcServer = SimpleXMLRPCServer((negoServer.ipAddress, int(negoServer.port)), allow_none=True, logRequests = False)
    logs.info("NEGO-SERVER: Listening on ipAddress " + negoServer.ipAddress + ", and port: "+str(negoServer.port)+" ...")
    rpcServer.register_instance(negoServer)
    negoServer.registerInDht()
    try:
        rpcServer.serve_forever()
    except:
        logs.trace(e)


