#! /usr/bin/env python

import time
import xmlrpclib
import random
import threading
import socket
import logs
import os
import time
from peerDescriptor import PeerDescriptor
from negoClient import NegoClient
from tManServer import PeerScore, peerScoreFromDict
from common import QueueWithTimeouts

class Reservation:
    def __init__(self, peer):
        self.peerId = peer.peerId
        self.score = peer.score
        self.size = peer.size
        self.timestamp = time.time()

    def timeouted(self):
        return time.time() - self.timestamp > 60 * 60 * 24 # 24 hours timeout


class TManClient(object):

    def __init__(self, dhtAddress, myPeerId, myUsername, myIp, myBasePort, config, filesLocationsManager):
        self.myPeerId = str(myPeerId)
        self.username = myUsername
        self.myIp = myIp
        self.myBasePort = myBasePort
        self.dhtAddress = dhtAddress
        self.dhtProxy = xmlrpclib.ServerProxy(dhtAddress, allow_none=True)
        self.filesLocationsManager = filesLocationsManager
        self.config = config
        self.waiting = QueueWithTimeouts()
        socket.setdefaulttimeout(self.config.socketTimeout)

    def startDaemon(self):
        lastInvokationForBalansing = time.time()
        lastInvokationForNego = time.time()
        lastInvokationForContractsVerification = time.time()
        lastInvokationForContractsRepair = time.time()
        lastInvokationForContractsCommit = time.time()

        while True:
            try:
                self.periodicWakeup()
                if (time.time() - lastInvokationForBalansing) > self.config.timeBetweenBalansingSec:
                    self.balansing()
                    lastInvokationForBalansing = time.time()
                if (time.time() - lastInvokationForNego) > self.config.negoServerIntervalSec:
                    self.wakeupNegoServer()
                    lastInvokationForNego = time.time()
                if (time.time() - lastInvokationForContractsVerification) > self.config.contractsVerificationIntervalSec:
                    self.verifyContracts()
                    lastInvokationForContractsVerification = time.time()
                if (time.time() - lastInvokationForContractsRepair) > self.config.contractsRepairIntervalSec:
                    self.repairInconsistentContracts()
                    lastInvokationForContractsRepair = time.time()
                if (time.time() - lastInvokationForContractsCommit) > self.config.contractsCommitIntervalSec:
                    self.tryCommitContracts()
                    lastInvokationForContractsVerification = time.time()
                time.sleep(self.config.tManSereverWakeupSec)
            except socket.error as e:
                logs.commonErr(e)
            except Exception as e:
                logs.trace(e)

    def verifyContracts(self):
        myNegoProxy = xmlrpclib.ServerProxy("http://"+str(self.myIp) + ":" + str(self.myBasePort), allow_none = True)
        myNegoProxy.verifyContracts()

    def repairInconsistentContracts(self):
        myNegoProxy = xmlrpclib.ServerProxy("http://"+str(self.myIp) + ":" + str(self.myBasePort), allow_none = True)
        myNegoProxy.repairInconsistentContracts()
    
    def tryCommitContracts(self):
        myTManProxy = xmlrpclib.ServerProxy("http://"+str(self.myIp) + ":" + str(int(self.myBasePort) + 2), allow_none = True)
        myTManProxy.tryCommitContracts()       

    def balansing(self):
        logs.debug("TMAN_CLIENT: Balansing loads")
        myNegoProxy = xmlrpclib.ServerProxy("http://"+str(self.myIp) + ":" + str(self.myBasePort), allow_none = True)
        freeStorage = self.filesLocationsManager.getTotalStorage() - int(myNegoProxy.getOldAgreementsSize()) - int(myNegoProxy.getAgreementsSize())
        myProxy = xmlrpclib.ServerProxy("http://"+str(self.myIp) + ":" + str(int(self.myBasePort) + 2), allow_none = True)
        myProxy.updateKnownPeers()
        peersByScore = [peerScoreFromDict(di) for di in myProxy.getPeersWithLowestScore()]

        for p in peersByScore:
            logs.debug("TMAN_CLIENT: Peers scores: " + str(p.score) + " - > " + str(p.peerId) + ", with version: " + str(p.version))

        for p in peersByScore:
            if (not self.waiting.has(p.peerId)) and (self.myPeerId != p.peerId) :
                if self.proposeReplication(p.peerId, freeStorage, myProxy, myNegoProxy):
                    break

    def proposeReplication(self, peerId, freeStorage, myProxy, myNegoProxy):
        logs.info("TMAN_CLIENT: proposing " + str(freeStorage) + " to " + str(peerId))
        try:
            remotePeer = PeerDescriptor(self.dhtAddress, str(peerId))
            remoteProxy = xmlrpclib.ServerProxy("http://"+str(remotePeer.ipAddress) + ":" + str(int(remotePeer.portNumber) + 2), allow_none = True)
            self.filesLocationsManager.preparePerfromanceInfo()
            ver = self.dhtProxy.getVersion(self.myPeerId)
            remoteProxy.updatePerformanceInfo(self.filesLocationsManager.perfInfo, ver)
            (accepted, changedPerfInfo) = remoteProxy.proposeReplication(self.myPeerId, self.username, str(freeStorage)) # the result is the size of accepted storage
            if not changedPerfInfo:
                if accepted:
                    self.waiting.add(peerId, self.config.timeToServePeerAfterAcceptanceOfProposalSec)
                    return True
                else:
                    self.waiting.add(peerId, self.config.timeToRepeatWhenRejectedContractProposalSec)
                    return False
        except socket.error as e:
            logs.commonErr(e)
            self.waiting.add(peerId, self.config.timeToRepeatWhenServerNotActiveSec)
            return False
        except Exception as e:
            logs.trace(e)
            self.waiting.add(peerId, self.config.timeToRepeatWhenServerNotActiveSec)
            return False


    def wakeupNegoServer(self):
        myNegoProxy = xmlrpclib.ServerProxy("http://"+str(self.myIp) + ":" + str(self.myBasePort), allow_none = True)
        myNegoProxy.wakeup()

    def periodicWakeup(self):
        self.updatePerfInfosForPeers()
        self.updateScore()
        myProxy = xmlrpclib.ServerProxy("http://"+str(self.myIp) + ":" + str(int(self.myBasePort) + 2), allow_none = True)
        sendList = [peerScoreFromDict(di) for di in myProxy.getPeersWithLowestScore()]
        randomList = [peerScoreFromDict(di) for di in myProxy.getRandomPeers()]
        myPeerScore = peerScoreFromDict(myProxy.getMyScore())
        if len(sendList + randomList) > 0:
            peerScore = random.choice(sendList + randomList)
            try:
                remotePeer = PeerDescriptor(self.dhtAddress, peerScore.peerId)
                remoteProxy = xmlrpclib.ServerProxy("http://"+str(remotePeer.ipAddress) + ":" + str(int(remotePeer.portNumber) + 2), allow_none = True)
            
                updateInfo = remoteProxy.updatePeers(sendList + randomList + [myPeerScore])
                updateInfo = [peerScoreFromDict(di) for di in updateInfo]
                randomPeers = remoteProxy.updateRandom(randomList + [myPeerScore])
                randomPeers = [peerScoreFromDict(di) for di in randomPeers]

                myProxy.updatePeers(updateInfo)
                myProxy.updateRandom(randomPeers)
            except socket.error as e:
                logs.commonErr(e)
            except Exception as e:
                logs.trace(e)
                logs.debug("Unable to send updateInfo to peer: " + str(peerScore.peerId))

    def updatePerfInfosForPeers(self):
        self.filesLocationsManager.preparePerfromanceInfo()

        try:
            myPeerDescriptor = PeerDescriptor(self.dhtAddress, self.myPeerId)
        except xmlrpclib.Fault:
            myPeerDescriptor = PeerDescriptor()

        self.dhtProxy.putField(str(self.myPeerId), "totalIncommingSize", self.filesLocationsManager.perfInfo.totalIncommingSize)
        self.dhtProxy.putField(str(self.myPeerId), "totalIncommingBwBound", self.filesLocationsManager.perfInfo.totalIncommingBwBound)

    # Here the scores of my data are updated
    def updateScore(self):
        self.filesLocationsManager.updateScores()
        if len(self.filesLocationsManager.scores) > 0:
            maxScore = -100000000000
            for fId, (score, size) in self.filesLocationsManager.scores.items():
                maxScore = max(maxScore, score)
            myProxy = xmlrpclib.ServerProxy("http://"+str(self.myIp) + ":" + str(int(self.myBasePort) + 2), allow_none = True)
            myProxy.updateMyScore(maxScore)


def startLoadBalansingDaemon(dhtAddress, myPeerId, myUsername, myIp, myBasePort, config,filesLocationsManager):
    tManClient = TManClient(dhtAddress, myPeerId, myUsername, myIp, myBasePort, config, filesLocationsManager)
    tManClient.startDaemon()

